/* _NVRM_COPYRIGHT_BEGIN_
 *
 * Copyright 2001 by NVIDIA Corporation.  All rights reserved.  All
 * information contained herein is proprietary and confidential to NVIDIA
 * Corporation.  Any use, reproduction, or disclosure without the written
 * permission of NVIDIA Corporation is prohibited.
 *
 * _NVRM_COPYRIGHT_END_
 */


#ifndef _NV_LINUX_H_
#define _NV_LINUX_H_

#include "nv.h"

#include <linux/config.h>
#include <linux/version.h>


#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 0)
#  error This driver does not support pre-2.4 kernels!
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 5, 0)
#  define KERNEL_2_4
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 0)
#  error This driver does not support 2.5 kernels!
#elif LINUX_VERSION_CODE < KERNEL_VERSION(2, 7, 0)
#  define KERNEL_2_6
#else
#  error This driver does not support development kernels!
#endif

#if defined (__ia64)
#  if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 13)
#    error This driver does not support 2.4.12 or earlier kernels!
#  endif
#endif

#if defined (CONFIG_SMP) && !defined (__SMP__)
#define __SMP__
#endif

#if defined (CONFIG_MODVERSIONS) && !defined (MODVERSIONS)
#  define MODVERSIONS
#endif

#if defined (MODVERSIONS) && !defined (KERNEL_2_6)
#include <linux/modversions.h>
#endif

#include <linux/kernel.h>
#include <linux/module.h>

#include <linux/init.h>             /* module_init, module_exit         */
#include <linux/types.h>            /* pic_t, size_t, __u32, etc        */
#include <linux/errno.h>            /* error codes                      */
#include <linux/list.h>             /* circular linked list             */
#include <linux/stddef.h>           /* NULL, offsetof                   */
#include <linux/wait.h>             /* wait queues                      */

#include <linux/slab.h>             /* kmalloc, kfree, etc              */
#include <linux/vmalloc.h>          /* vmalloc, vfree, etc              */

#include <linux/poll.h>             /* poll_wait                        */
#include <linux/delay.h>            /* mdelay, udelay                   */

#ifdef KERNEL_2_6
#include <linux/sched.h>            /* suser(), capable() replacement   */
#include <linux/moduleparam.h>      /* module_param()                   */
#include <linux/smp_lock.h>         /* kernel_locked                    */
#include <asm/tlbflush.h>           /* flush_tlb(), flush_tlb_all()     */
#include <asm/kmap_types.h>         /* page table entry lookup          */
#endif

#include <linux/pci.h>              /* pci_find_class, etc              */
#include <linux/interrupt.h>        /* tasklets, interrupt helpers      */
#include <linux/timer.h>

#include <asm/system.h>             /* cli, sli, save_flags             */
#include <asm/io.h>                 /* ioremap, virt_to_phys            */
#include <asm/uaccess.h>            /* access_ok                        */
#include <asm/page.h>               /* PAGE_OFFSET                      */
#include <asm/pgtable.h>            /* pte bit definitions              */

#if defined(NVCPU_X86_64) && defined(KERNEL_2_6)
#include <linux/syscalls.h>         /* sys_ioctl()                      */
#include <linux/ioctl32.h>          /* register_ioctl32_conversion()    */
#endif

#if defined(NVCPU_X86_64) && defined(KERNEL_2_4)
#include <asm/ioctl32.h>            /* sys_ioctl() (ioctl32)            */
#endif

#include <linux/spinlock.h>
#include <asm/semaphore.h>
#include <linux/highmem.h>

#ifdef CONFIG_PROC_FS
#include <linux/proc_fs.h>
#endif

#ifdef CONFIG_DEVFS_FS
#include <linux/devfs_fs_kernel.h>
#endif

#ifdef CONFIG_KMOD
#include <linux/kmod.h>
#endif

#ifdef CONFIG_PM
#include <linux/pm.h>
#endif

#ifdef CONFIG_MTRR
#include <asm/mtrr.h>
#endif

#ifdef CONFIG_KDB
#include <linux/kdb.h>
#include <asm/kdb.h>
#endif

#if defined (CONFIG_AGP) || defined (CONFIG_AGP_MODULE)
#ifndef NOAGPGART
#  define AGPGART
#  include <linux/agp_backend.h>
#  include <linux/agpgart.h>
#endif
#endif

#if defined(NVCPU_X86) || defined(NVCPU_X86_64)
#define NV_BUILD_NV_PAT_SUPPORT 1
#endif

#if defined(NV_BUILD_NV_PAT_SUPPORT)
#include "pat.h"
#endif

#if defined(NVCPU_X86)
#ifndef write_cr4
#define write_cr4(x) __asm__ ("movl %0,%%cr4" :: "r" (x));
#endif

#ifndef read_cr4
#define read_cr4()                                  \
 ({                                                 \
      unsigned int __cr4;                           \
      __asm__ ("movl %%cr4,%0" : "=r" (__cr4));     \
      __cr4;                                        \
  })
#endif

#ifndef wbinvd
#define wbinvd() __asm__ __volatile__("wbinvd" ::: "memory");
#endif
#endif /* defined(NVCPU_X86) */


#if !defined (list_for_each)
#define list_for_each(pos, head) \
        for (pos = (head)->next; pos != (head); pos = (pos)->next)
#endif

#ifdef EXPORT_NO_SYMBOLS
/* don't clutter the kernel namespace with our symbols */
EXPORT_NO_SYMBOLS;
#endif

#if !defined(DEBUG) && defined(__GFP_NOWARN)
#define NV_GFP_KERNEL (GFP_KERNEL | __GFP_NOWARN)
#define NV_GFP_ATOMIC (GFP_ATOMIC | __GFP_NOWARN)
#else
#define NV_GFP_KERNEL (GFP_KERNEL)
#define NV_GFP_ATOMIC (GFP_ATOMIC)
#endif

#if defined(NVCPU_IA64)
#define NV_GFP_DMA32 (NV_GFP_KERNEL | __GFP_DMA)
#define CACHE_FLUSH()
#else
#define NV_GFP_DMA32 (NV_GFP_KERNEL)
#define CACHE_FLUSH()  asm volatile("wbinvd":::"memory")
#endif

#ifndef NVWATCH

/* various memory tracking/debugging techniques
 * disabled for retail builds, enabled for debug builds
 */

// allow an easy way to convert all debug printfs related to memory
// management back and forth between 'info' and 'errors'
#if defined(NV_DBG_MEM)
#define NV_DBG_MEMINFO NV_DBG_ERRORS
#else
#define NV_DBG_MEMINFO NV_DBG_INFO
#endif

#ifdef DEBUG
#define NV_ENABLE_MEM_TRACKING 1
#endif

#if NV_ENABLE_MEM_TRACKING
#define NV_MEM_TRACKING_PAD_SIZE(size)   ((size) += sizeof(void *))
#define NV_MEM_TRACKING_HIDE_SIZE(ptr, size)            \
    if ((ptr) && *(ptr)) {                              \
        U008 *__ptr;                                    \
        *(unsigned long *) *(ptr) = (size);             \
        __ptr = *(ptr); __ptr += sizeof(void *);        \
        *(ptr) = (void *) __ptr;                        \
    }
#define NV_MEM_TRACKING_RETRIEVE_SIZE(ptr, size)        \
    {                                                   \
        U008 *__ptr = (ptr); __ptr -= sizeof(void *);   \
        (ptr) = (void *) __ptr;                         \
        size = *(unsigned long *) (ptr);                \
    }
#else
#define NV_MEM_TRACKING_PAD_SIZE(size)
#define NV_MEM_TRACKING_HIDE_SIZE(ptr, size)
#define NV_MEM_TRACKING_RETRIEVE_SIZE(ptr, size)  ((size) = 0)
#endif


/* poor man's memory allocation tracker.
 * main intention is just to see how much memory is being used to recognize
 * when memory usage gets out of control or if memory leaks are happening
 */

/* keep track of memory usage */
#if NV_ENABLE_MEM_TRACKING
extern int vm_usage;
extern int km_usage;
extern int fp_usage;

/* print out a running tally of memory allocation amounts, disabled by default */
// #define POOR_MANS_MEM_CHECK 1


/* slightly more advanced memory allocation tracker.
 * track who's allocating memory and print out a list of currently allocated
 * memory at key points in the driver
 */

#define MEMDBG_ALLOC(a,b) (a = kmalloc(b, NV_GFP_ATOMIC))
#define MEMDBG_FREE(a)    (kfree(a))

#include "nv-memdbg.h"

#undef MEMDBG_ALLOC
#undef MEMDBG_FREE

extern struct mem_track_t *vm_list;
extern struct mem_track_t *km_list;
extern struct mem_track_t *fp_list;

/* print out list of memory allocations */
/* default to enabled for now */
#define LIST_MEM_CHECK 1

/* decide which memory types to apply mem trackers to */
#define VM_CHECKER 1
#define KM_CHECKER 1
#define FP_CHECKER 1

#endif  /* NV_ENABLE_MEM_TRACKING */

#define VM_PRINT(str, args...)
#define KM_PRINT(str, args...)
#define FP_PRINT(str, args...)

#define VM_ADD_MEM(a,b,c,d)
#define VM_FREE_MEM(a,b,c,d)
#define KM_ADD_MEM(a,b,c,d)
#define KM_FREE_MEM(a,b,c,d)
#define FP_ADD_MEM(a,b,c,d)
#define FP_FREE_MEM(a,b,c,d)

#define VM_ALLOC_RECORD(a,b,c)
#define VM_FREE_RECORD(a,b,c)
#define KM_ALLOC_RECORD(a,b,c)
#define KM_FREE_RECORD(a,b,c)


#if defined(VM_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  VM_PRINT
#    define VM_PRINT(str, args...)   printk(str, ##args)
#  endif
#  if defined(LIST_MEM_CHECK)
#    undef  VM_ADD_MEM
#    define VM_ADD_MEM(a,b,c,d)      nv_add_mem(&vm_list, a, b, c, d)
#    undef  VM_FREE_MEM
#    define VM_FREE_MEM(a,b,c,d)     nv_free_mem(&vm_list, a, b, c, d)
#  endif
#  undef VM_ALLOC_RECORD
#  define VM_ALLOC_RECORD(ptr, size, name) \
      if (ptr) { \
            vm_usage += size; \
            VM_PRINT("%s (0x%x: 0x%x): vm_usage is now 0x%x\n", \
                name, ptr, size, vm_usage); \
            VM_ADD_MEM(ptr, size, __FILE__, __LINE__); \
        }
#  undef VM_FREE_RECORD
#  define VM_FREE_RECORD(ptr, size, name) \
        if (ptr) { \
            vm_usage -= size; \
            VM_PRINT("%s (0x%x: 0x%x): vm_usage is now 0x%x\n", \
                name, ptr, size, vm_usage); \
            VM_FREE_MEM(ptr, size, __FILE__, __LINE__); \
        }
#endif

#if defined(KM_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  KM_PRINT
#    define KM_PRINT(str, args...)   printk(str, ##args)
#  endif
#  if defined(LIST_MEM_CHECK)
#    undef  KM_ADD_MEM
#    define KM_ADD_MEM(a,b,c,d)      nv_add_mem(&km_list, a, b, c, d)
#    undef  KM_FREE_MEM
#    define KM_FREE_MEM(a,b,c,d)     nv_free_mem(&km_list, a, b, c, d)
#  endif
#  undef KM_ALLOC_RECORD
#  define KM_ALLOC_RECORD(ptr, size, name) \
      if (ptr) { \
            km_usage += size; \
            KM_PRINT("%s (0x%x: 0x%x): km_usage is now 0x%x\n", \
                name, ptr, size, km_usage); \
            KM_ADD_MEM(ptr, size, __FILE__, __LINE__); \
        }
#  undef KM_FREE_RECORD
#  define KM_FREE_RECORD(ptr, size, name) \
        if (ptr) { \
            km_usage -= size; \
            KM_PRINT("%s (0x%x: 0x%x): km_usage is now 0x%x\n", \
                name, ptr, size, km_usage); \
            KM_FREE_MEM(ptr, size, __FILE__, __LINE__); \
        }
#endif

#if defined(FP_CHECKER)
#  if defined(POOR_MANS_MEM_CHECK)
#    undef  FP_PRINT
#    define FP_PRINT(str, args...)   printk(str, ##args)
#  endif
#if 0  // I had some problems tracking fp mem, so disable for now
#  if defined(LIST_MEM_CHECK)
#    undef  FP_ADD_MEM
#    define FP_ADD_MEM(a,b,c,d)      nv_add_mem(&fp_list, a, b, c, d)
#    undef  FP_FREE_MEM
#    define FP_FREE_MEM(a,b,c,d)     nv_free_mem(&fp_list, a, b, c, d)
#  endif
#endif
#endif

/* NV_VMALLOC has to allocate memory under the 4GB mark suitable for
 * DMA use with the hardware.  This has to be done differently for
 * different platforms and kernel versions.
 */
#if defined(NVCPU_IA64)
#define NV_VMALLOC(ptr, size) \
    { \
        (ptr) = vmalloc_dma(size); \
        VM_ALLOC_RECORD(ptr, size, "vm_alloc"); \
    }
#else
#define NV_VMALLOC(ptr, size) \
    { \
        (ptr) = vmalloc_32(size); \
        VM_ALLOC_RECORD(ptr, size, "vm_alloc"); \
    }
#endif

#define NV_VFREE(ptr, size) \
    { \
        VM_FREE_RECORD(ptr, size, "vm_alloc"); \
        vfree((void *) (ptr)); \
    }

#define NV_IOREMAP(ptr, physaddr, size) \
    { \
        (ptr) = ioremap(physaddr, size); \
        VM_ALLOC_RECORD(ptr, size, "vm_ioremap"); \
    }

#define NV_IOREMAP_NOCACHE(ptr, physaddr, size) \
    { \
        (ptr) = ioremap_nocache(physaddr, size); \
        VM_ALLOC_RECORD(ptr, size, "vm_ioremap_nocache"); \
    }

#define NV_IOUNMAP(ptr, size) \
    { \
        VM_FREE_RECORD(ptr, size, "vm_iounmap"); \
        iounmap(ptr); \
    }

/* only use this because GFP_KERNEL may sleep..
 * GFP_ATOMIC is ok, it won't sleep
 */
#define NV_KMALLOC(ptr, size) \
    { \
        (ptr) = kmalloc(size, NV_GFP_KERNEL); \
        KM_ALLOC_RECORD(ptr, size, "km_alloc"); \
    }

#define NV_KMALLOC_ATOMIC(ptr, size) \
    { \
        (ptr) = kmalloc(size, NV_GFP_ATOMIC); \
        KM_ALLOC_RECORD(ptr, size, "km_alloc_atomic"); \
    }  


#define NV_KFREE(ptr, size) \
    { \
        KM_FREE_RECORD(ptr, size, "km_free"); \
        kfree((void *) (ptr)); \
    }

#define NV_GET_FREE_PAGES(ptr, order) \
    { \
        (ptr) = __get_free_pages(NV_GFP_DMA32, order); \
    }
        
#define NV_FREE_PAGES(ptr, order) \
    { \
        free_pages(ptr, order); \
    }

#endif /* !defined NVWATCH */


#if defined (KERNEL_2_4)
#  define NV_IS_SUSER()                 suser()
#  define NV_PCI_DEVICE_NAME(dev)       ((dev)->name)
#  define NV_NUM_CPUS()                 smp_num_cpus
#  define NV_CLI()                      __cli()
#  define NV_SAVE_FLAGS(eflags)         __save_flags(eflags)
#  define NV_RESTORE_FLAGS(eflags)      __restore_flags(eflags)
#  define NV_MAY_SLEEP()                (!in_interrupt())
#  define NV_MODULE_PARAMETER(x)        MODULE_PARM(x, "i")
#endif

#if defined (KERNEL_2_6)
#  define NV_IS_SUSER()                 capable(CAP_SYS_ADMIN)
#  define NV_PCI_DEVICE_NAME(dev)       ((dev)->pretty_name)
#  define NV_NUM_CPUS()                 num_online_cpus()
#  define NV_CLI()                      local_irq_disable()
#  define NV_SAVE_FLAGS(eflags)         local_save_flags(eflags)
#  define NV_RESTORE_FLAGS(eflags)      local_irq_restore(eflags)
#  define NV_MAY_SLEEP()                (!in_interrupt() && !in_atomic())
#  define NV_MODULE_PARAMETER(x)        module_param(x, int, 0)

   // the following macro causes problems when used in the same module
   // as module_param(); undef it so we don't accidentally mix the two
#  undef  MODULE_PARM
#endif

   // Earlier 2.4.x kernels don't have pci_disable_device()
#ifdef NV_PCI_DISABLE_DEVICE_PRESENT
#define NV_PCI_DISABLE_DEVICE(dev)      pci_disable_device(dev)
#else
#define NV_PCI_DISABLE_DEVICE(dev)
#endif

/* common defines */
#define GET_MODULE_SYMBOL(mod,sym)    (const void *) inter_module_get(sym)
#define PUT_MODULE_SYMBOL(sym)        inter_module_put((char *) sym)

#define NV_GET_PAGE_STRUCT(phys_page) virt_to_page(__va(phys_page))
#define NV_VMA_OFFSET(vma)            (((vma)->vm_pgoff) << PAGE_SHIFT)
#define NV_VMA_PRIVATE(vma)           ((vma)->vm_private_data)

#define NV_DEVICE_NUMBER(x)           minor((x)->i_rdev)
#define NV_IS_CONTROL_DEVICE(x)       (minor((x)->i_rdev) == 255)

#define NV_PCI_RESOURCE_START(dev, bar) ((dev)->resource[(bar) - 1].start)
#define NV_PCI_RESOURCE_SIZE(dev, bar)  ((dev)->resource[(bar) - 1].end - (dev)->resource[(bar) - 1].start + 1)

#define NV_PCI_BUS_NUMBER(dev)        (dev)->bus->number
#define NV_PCI_SLOT_NUMBER(dev)       PCI_SLOT((dev)->devfn)

#ifdef NV_PCI_GET_CLASS_PRESENT
#define NV_PCI_DEV_PUT(dev)                    pci_dev_put(dev)
#define NV_PCI_GET_DEVICE(vendor,device,from)  pci_get_device(vendor,device,from)
#define NV_PCI_GET_SLOT(bus,devfn)             pci_get_slot(pci_find_bus(0,bus),devfn)
#define NV_PCI_GET_CLASS(class,from)           pci_get_class(class,from)
#else
#define NV_PCI_DEV_PUT(dev)
#define NV_PCI_GET_DEVICE(vendor,device,from)  pci_find_device(vendor,device,from)
#define NV_PCI_GET_SLOT(bus,devfn)             pci_find_slot(bus,devfn)
#define NV_PCI_GET_CLASS(class,from)           pci_find_class(class,from)
#endif

#if defined(DEBUG)
#define NV_PRINT_AT(at)                                                    \
    nv_printf(NV_DBG_ERRORS, "nvidia lost alloc_t: %4d  %2d  %05x 0x%08x 0x%08x\n", \
        at->num_pages, at->usage_count, \
        at->flags, at->page_table, at->key_mapping)
#else
#define NV_PRINT_AT(at)
#endif

// acpi support has been back-ported to the 2.4 kernel, but the 2.4 driver
// model is not sufficient for full acpi support. it may work in some cases,
// but not enough for us to officially support this configuration.
#if defined(CONFIG_ACPI) && defined(KERNEL_2_6)
#define NV_PM_SUPPORT_ACPI
#endif

#if defined(CONFIG_APM) || defined(CONFIG_APM_MODULE)
#define NV_PM_SUPPORT_APM
#endif

/* add support for iommu.
 * on x86_64 platforms, this uses the gart to remap pages that are > 32-bits
 * to < 32-bits.
 */
#if defined(NVCPU_X86_64)
#define NV_SG_MAP_BUFFERS 1
#endif

/* add support for software i/o tlb support.
 * normally, you'd expect this to be transparent, but unfortunately this is not
 * the case. for starters, the sw io tlb is a pool of pre-allocated pages that
 * are < 32-bits. when we ask to remap a page through this sw io tlb, we are
 * returned one of these pages, which means we have 2 different pages, rather
 * than 2 mappings to the same page. secondly, this pre-allocated pool is very
 * tiny, and the kernel panics when it is exhausted. try to warn the user that
 * they need to boost the size of their pool.
 */
#if defined(CONFIG_SWIOTLB)
#define NV_SWIOTLB 1
extern int nv_swiotlb;
#endif

/*
 * early 2.6 kernels changed their swiotlb codepath, running into a
 * latent bug that returns virtual addresses when it should return
 * physical addresses. we try to gracefully account for that, by 
 * comparing the returned address to what should be it's virtual
 * equivalent. this should hopefully account for when the bug is 
 * fixed in the core kernel.
 */
#if defined(NV_SWIOTLB) && defined(KERNEL_2_6)
#define NV_FIXUP_SWIOTLB_VIRT_ADDR_BUG(dma_addr) \
    if ((dma_addr) == ((dma_addr) | PAGE_OFFSET)) \
        (dma_addr) = __pa((dma_addr))
#else
#define NV_FIXUP_SWIOTLB_VIRT_ADDR_BUG(dma_addr)
#endif

#ifndef minor
# define minor(x) MINOR(x)
#endif

#ifndef IRQ_HANDLED
typedef void irqreturn_t;
#define IRQ_HANDLED
#endif

#ifndef PCI_CAP_ID_EXP
#define PCI_CAP_ID_EXP 0x10
#endif

#if defined(KERNEL_2_6) && defined(AGPGART)
typedef struct agp_kern_info agp_kern_info;
typedef struct agp_memory agp_memory;
#endif

#if defined(CONFIG_DEVFS_FS)
#  if defined(KERNEL_2_6)
     typedef void* devfs_handle_t;
#    define NV_DEVFS_REGISTER(_name, _minor)                             \
     ({                                                                  \
         devfs_handle_t __handle = NULL;                                 \
         if (devfs_mk_cdev(MKDEV(NV_MAJOR_DEVICE_NUMBER, _minor),        \
                 S_IFCHR | S_IRUGO | S_IWUGO, _name) == 0)               \
            __handle = (void *) 1; /* XXX Fix me! (boolean) */           \
         __handle;                                                       \
     })

#    define NV_DEVFS_REMOVE_DEVICE(i) devfs_remove("nvidia%d", i)
#    define NV_DEVFS_REMOVE_CONTROL() devfs_remove("nvidiactl")
#  else // defined(KERNEL_2_4)
#    define NV_DEVFS_REGISTER(_name, _minor)                             \
     ({                                                                  \
         devfs_handle_t __handle =                                       \
             devfs_register(NULL, _name, DEVFS_FL_DEFAULT,               \
                     NV_MAJOR_DEVICE_NUMBER, _minor,                     \
                     S_IFCHR | S_IRUGO | S_IWUGO, &nv_fops, NULL);       \
         __handle;                                                       \
     })

#    define NV_DEVFS_REMOVE_DEVICE(i)                                    \
     ({                                                                  \
         if (nv_devfs_handles[i+1] != NULL)                              \
             devfs_unregister(nv_devfs_handles[i+1]);                    \
      })
#    define NV_DEVFS_REMOVE_CONTROL()                                    \
     ({                                                                  \
         if (nv_devfs_handles[0] != NULL)                                \
             devfs_unregister(nv_devfs_handles[0]);                      \
      })
#  endif // defined(KERNEL_2_4)
#endif // defined(CONFIG_DEVFS_FS)

#if defined(CONFIG_DEVFS_FS) && !defined(KERNEL_2_6)
#define NV_REGISTER_CHRDEV(x...)    devfs_register_chrdev(x)
#define NV_UNREGISTER_CHRDEV(x...)  devfs_unregister_chrdev(x)
#else
#define NV_REGISTER_CHRDEV(x...)    register_chrdev(x)
#define NV_UNREGISTER_CHRDEV(x...)  unregister_chrdev(x)
#endif

#if defined(NV_REMAP_PFN_RANGE_PRESENT)
#define NV_REMAP_PAGE_RANGE(from, offset, x...) \
     remap_pfn_range(vma, from, ((offset) >> PAGE_SHIFT), x)
#elif defined(NV_REMAP_PAGE_RANGE_5)
#define NV_REMAP_PAGE_RANGE(x...) remap_page_range(vma, x)
#elif defined(NV_REMAP_PAGE_RANGE_4)
#define NV_REMAP_PAGE_RANGE(x...) remap_page_range(x)
#else
#warning "conftest.sh failed, assuming remap_page_range(4)!"
#define NV_REMAP_PAGE_RANGE(x...) remap_page_range(x)
#endif

#if defined(pmd_offset_map)
#define NV_PMD_OFFSET(addres, pg_dir, pg_mid_dir) \
    { \
        pg_mid_dir = pmd_offset_map(pg_dir, address); \
    }
#define NV_PMD_UNMAP(pg_mid_dir) \
    { \
        pmd_unmap(pg_mid_dir); \
    }
#else
#define NV_PMD_OFFSET(addres, pg_dir, pg_mid_dir) \
    { \
        pg_mid_dir = pmd_offset(pg_dir, address); \
    }
#define NV_PMD_UNMAP(pg_mid_dir)
#endif

#define NV_PMD_PRESENT(pg_mid_dir) \
    ({ \
        if ( (pg_mid_dir) && (pmd_none(*pg_mid_dir))) { \
            NV_PMD_UNMAP(pg_mid_dir); pg_mid_dir = NULL; \
        } pg_mid_dir != NULL; \
    })

#if defined(pte_offset_atomic)
#define NV_PTE_OFFSET(addres, pg_mid_dir, pte) \
    { \
        pte = pte_offset_atomic(pg_mid_dir, address); \
        NV_PMD_UNMAP(pg_mid_dir); \
    }
#define NV_PTE_UNMAP(pte) \
    { \
        pte_kunmap(pte); \
    }
#elif defined(pte_offset)
#define NV_PTE_OFFSET(addres, pg_mid_dir, pte) \
    { \
        pte = pte_offset(pg_mid_dir, address); \
        NV_PMD_UNMAP(pg_mid_dir); \
    }
#define NV_PTE_UNMAP(pte)
#else
#define NV_PTE_OFFSET(addres, pg_mid_dir, pte) \
    { \
        pte = pte_offset_map(pg_mid_dir, address); \
        NV_PMD_UNMAP(pg_mid_dir); \
    }
#define NV_PTE_UNMAP(pte) \
    { \
        pte_unmap(pte); \
    }
#endif

#define NV_PTE_PRESENT(pte) \
    ({ \
        if (pte) { \
            if (!pte_present(*pte)) { \
                NV_PTE_UNMAP(pte); pte = NULL; \
            } \
        } pte != NULL; \
    })

#define NV_PTE_VALUE(pte) \
    ({ \
        unsigned long __pte_value = pte_val(*pte); \
        NV_PTE_UNMAP(pte); \
        __pte_value; \
    })

#define NV_PAGE_ALIGN(addr)             ( ((addr) + PAGE_SIZE - 1) / PAGE_SIZE)
#define NV_MASK_OFFSET(addr)            ( (addr) & (PAGE_SIZE - 1) )

#if defined(NVCPU_X86) || defined(NVCPU_X86_64)
/* this isn't defined in some older kernel header files */
#define NV_CPU_INTERRUPT_FLAGS_BIT (1<<9)
#elif defined(NVCPU_IA64)
/* For whatever reason this is not defined an any header file I could
 * find.  From Intel IA64 Architecture Software Developers Manual Volume 2: 
 * IA64 System Architecture page 3-7 we have:
 */
#define NV_CPU_INTERRUPT_FLAGS_BIT (1<<14)
#else
#error define NV_CPU_INTERRUPT_FLAGS_BIT
#endif

static inline int NV_IRQL_IS_RAISED(void)
    {
        unsigned long int eflags;
        NV_SAVE_FLAGS(eflags);
        return !(eflags & NV_CPU_INTERRUPT_FLAGS_BIT);
    }
 
static inline int nv_calc_order(int size)
    {
        int order = 0;
        while ( ((1 << order) * PAGE_SIZE) < (size))
        {
            order++;
        }
        return order;
    }

#if !defined (pgprot_noncached)
static inline pgprot_t pgprot_noncached(pgprot_t old_prot)
    {
        pgprot_t new_prot = old_prot;
        if (boot_cpu_data.x86 > 3)
            new_prot = __pgprot(pgprot_val(old_prot) | _PAGE_PCD);
        return new_prot;
    }
#endif

#if defined(NV_BUILD_NV_PAT_SUPPORT) && !defined (pgprot_writecombined)
static inline pgprot_t pgprot_writecombined(pgprot_t old_prot)
    {
        pgprot_t new_prot = old_prot;
        if (boot_cpu_data.x86 > 3)
        {
            pgprot_val(old_prot) &= ~(_PAGE_PCD | _PAGE_PWT);
            new_prot = __pgprot(pgprot_val(old_prot) | _PAGE_WRTCOMB);
        }
        return new_prot;
    }
#endif

#if !defined(page_to_pfn)
#define page_to_pfn(page)  ((page) - mem_map)
#endif

/* On IA64 physical memory is partitioned into a cached and an
 * uncached view controlled by bit 63.  Set this bit when remapping
 * page ranges.  
 */
#if defined(NVCPU_IA64)
#define phys_to_uncached(addr) ((addr) | ((unsigned long) 1<<63))
#else
/* Some other scheme must be used on this platform */
#define phys_to_uncached(addr) (addr)
#endif

/*
 * An allocated bit of memory using NV_MEMORY_ALLOCATION_OFFSET
 *   looks like this in the driver
 */

typedef struct nv_pte_t {
    unsigned long   phys_addr;
    unsigned long   virt_addr;
    dma_addr_t      dma_addr;
#ifdef CONFIG_SWIOTLB
    unsigned long   orig_phys_addr;
    unsigned long   orig_virt_addr;
#endif
} nv_pte_t;

typedef struct nv_alloc_s {
    struct nv_alloc_s *next;    
    atomic_t       usage_count;
    unsigned int   flags;
    unsigned int   num_pages;
    unsigned int   order;
    unsigned int   size;
    nv_pte_t      *page_table;          /* list of physical pages allocated */
    void          *key_mapping;         /* mapping used as a key for finding this nv_alloc_t */
                                        /*   may be the same as page_table                   */
    unsigned int   class;
    void          *priv_data;
#if defined(NV_SG_MAP_BUFFERS)
    struct pci_dev *dev;
    struct scatterlist *sg_list;        /* list of physical pages allocated */
#endif
} nv_alloc_t;


#define NV_ALLOC_TYPE_PCI      (1<<0)
#define NV_ALLOC_TYPE_AGP      (1<<1)
#define NV_ALLOC_TYPE_CONTIG   (1<<2)
#define NV_ALLOC_TYPE_KERNEL   (1<<3)
#define NV_ALLOC_TYPE_VMALLOC  (1<<4)

#define NV_ALLOC_MAPPING_SHIFT      16
#define NV_ALLOC_MAPPING(flags)     (((flags)>>NV_ALLOC_MAPPING_SHIFT)&0xff)
#define NV_ALLOC_ENC_MAPPING(flags) ((flags)<<NV_ALLOC_MAPPING_SHIFT)

#define NV_ALLOC_MAPPING_CACHED(flags) ((NV_ALLOC_MAPPING(flags) == NV_MEMORY_DEFAULT) || \
                                        (NV_ALLOC_MAPPING(flags) == NV_MEMORY_WRITEBACK))

#define NV_ALLOC_MAPPING_CONTIG(flags) ((flags) & NV_ALLOC_TYPE_CONTIG)
#define NV_ALLOC_MAPPING_VMALLOC(flags) ((flags) & NV_ALLOC_TYPE_VMALLOC)

static inline U032 nv_alloc_init_flags(int cached, int agp, int contig, int kernel)
{
    U032 flags = NV_ALLOC_ENC_MAPPING(cached);
    if (agp)    flags |= NV_ALLOC_TYPE_AGP;
    else        flags |= NV_ALLOC_TYPE_PCI;
    if (kernel) flags |= NV_ALLOC_TYPE_KERNEL; 
    if (kernel && !contig) flags |= NV_ALLOC_TYPE_VMALLOC;
    if (contig && !agp) flags |= NV_ALLOC_TYPE_CONTIG;
    return flags;
}

/* linux-specific version of old nv_state_t */
/* this is a general os-specific state structure. the first element *must* be
   the general state structure, for the generic unix-based code */
typedef struct {
    nv_state_t nv_state;
    atomic_t usage_count;

    struct pci_dev *dev;
    nv_alloc_t *alloc_queue;

    /* keep track of any pending bottom halfes */
    struct tasklet_struct tasklet;

    /* get a timer callback every second */
    struct timer_list rc_timer;

    /* per-device locking mechanism for access to core rm */
    spinlock_t rm_lock;
    int rm_lock_cpu;
    int rm_lock_count;

    /* lock for linux-specific data, not used by core rm */
    struct semaphore ldata_lock;

    /* lock for linux-specific alloc queue */
    struct semaphore at_lock;
} nv_linux_state_t;


/*
 * file-private data
 * hide a pointer to our data structures in a file-private ptr
 * there are times we need to grab this data back from the file
 * data structure..
 */

#define NV_EVENT_FIFO_SIZE 6

typedef struct
{
    void *nvptr;
    U032 num_events;
    U032 put, get;
    spinlock_t fp_lock;
    wait_queue_head_t waitqueue;
    nv_event_t *event_fifo;     // fifo for storing events
} nv_file_private_t;

#define FILE_PRIVATE(filep)     ((filep)->private_data)

#define NV_GET_NVFP(filep)      ((nv_file_private_t *) FILE_PRIVATE(filep))

/* for the card devices */
#define NVL_FROM_FILEP(filep)   (NV_GET_NVFP(filep)->nvptr)

#define NV_GET_NVL_FROM_NV_STATE(nv) \
    ((nv_linux_state_t *) nv->os_state)

#define NV_STATE_PTR(nvl)   (&((nvl)->nv_state))


#define NV_ATOMIC_SET(data,val)         atomic_set(&(data), (val))
#define NV_ATOMIC_INC(data)             atomic_inc(&(data))
#define NV_ATOMIC_DEC(data)             atomic_dec(&(data))
#define NV_ATOMIC_DEC_AND_TEST(data)    atomic_dec_and_test(&(data))
#define NV_ATOMIC_READ(data)            atomic_read(&(data))

/*
 * AMD Athlon processors expose a subtle bug in the Linux
 * kernel, that may lead to AGP memory corruption. Recent
 * kernel versions had a workaround for this problem, but
 * 2.4.20 is the first kernel to address it properly. The
 * page_attr API provides the means to solve the problem. 
 */
#if defined(NVCPU_X86) && defined(NV_CHANGE_PAGE_ATTR_PRESENT)
static inline void NV_SET_PAGE_ATTRIB_UNCACHED(nv_pte_t *page_ptr)
    {
        struct page *page = virt_to_page(__va(page_ptr->phys_addr));
        change_page_attr(page, 1, PAGE_KERNEL_NOCACHE);
    }
static inline void NV_SET_PAGE_ATTRIB_CACHED(nv_pte_t *page_ptr)
    {
        struct page *page = virt_to_page(__va(page_ptr->phys_addr));
#ifdef CONFIG_SWIOTLB
        if (page_ptr->orig_phys_addr)
        {
            nv_printf(NV_DBG_ERRORS, "NVRM: trying to set page attrib on wrong page!\n");
            os_dbg_breakpoint();
        }
#endif
        change_page_attr(page, 1, PAGE_KERNEL);
    }
#else
#define NV_SET_PAGE_ATTRIB_UNCACHED(page_list)
#define NV_SET_PAGE_ATTRIB_CACHED(page_list)
#endif

static inline void NV_UNLOCK_PAGE(nv_pte_t *page_ptr)
    {
        unsigned long phys_addr;
#ifdef CONFIG_SWIOTLB
    	if (page_ptr->orig_phys_addr) phys_addr = page_ptr->orig_phys_addr;
        else
#endif
    	    phys_addr = page_ptr->phys_addr;
        ClearPageReserved(NV_GET_PAGE_STRUCT(phys_addr));
    }

#if defined(NV_SG_MAP_BUFFERS)
static inline int nv_sg_map_buffer(
    struct pci_dev     *dev,
    struct scatterlist *sg_ptr,
    void               *base,
    unsigned int        num_pages
)
{
    int i;

    sg_ptr->page = virt_to_page(base);
    sg_ptr->offset = (unsigned long)base & ~PAGE_MASK;
    sg_ptr->length  = num_pages * PAGE_SIZE;

    if (pci_map_sg(dev, sg_ptr, 1, PCI_DMA_BIDIRECTIONAL) == 0)
    {
        return 1;
    }

    NV_FIXUP_SWIOTLB_VIRT_ADDR_BUG(sg_ptr->dma_address);

    // this is a bit of a hack to make contiguous allocations easier to handle
    // nv_sg_load below relies on the page_ptr addresses being filed in, as 
    // well as the sg_ptr having a valid dma_address. most allocations call
    // nv_sg_map_buffers page-by-page, but contiguous allocations will make
    // one call for the whole allocation. make sure we correctly propogate 
    // our dma_address through the rest of the sg_ptrs for these allocations.
    // note we start with index 1, since index 0 is already correct
    for (i = 1; i < num_pages; i++)
    {
        sg_ptr[i].dma_address = sg_ptr[0].dma_address + (i * PAGE_SIZE);
    }

    return 0;
}

static inline int nv_sg_load(
    struct scatterlist *sg_ptr,
    nv_pte_t           *page_ptr
)
{
    page_ptr->dma_addr = sg_ptr->dma_address;

#if defined(NV_SWIOTLB)
    // with the sw io tlb, we've actually switched to different physical pages
    // wire in the new page's addresses, but save the original off to free later
    if (nv_swiotlb)
    {
        page_ptr->orig_phys_addr = page_ptr->phys_addr;
        page_ptr->phys_addr      = page_ptr->dma_addr;
        page_ptr->orig_virt_addr = page_ptr->virt_addr;
        page_ptr->virt_addr      = (unsigned long) __va(page_ptr->dma_addr);
    }
#endif

    return 0;
}

// make sure we only unmap the page if it was really mapped through the iommu,
// in which case the dma_addr and phys_addr will not match.
static inline void nv_sg_unmap_buffer(
    struct pci_dev     *dev,
    struct scatterlist *sg_ptr,
    nv_pte_t           *page_ptr
)
{
#ifdef CONFIG_SWIOTLB
    // for sw io tlbs, dma_addr == phys_addr currently, so the check below fails
    // restore the original settings first, then the following check will work
    if (nv_swiotlb && page_ptr->orig_phys_addr)
    {
        page_ptr->phys_addr      = page_ptr->orig_phys_addr;
        page_ptr->virt_addr      = page_ptr->orig_virt_addr;
        page_ptr->orig_phys_addr = 0;
        page_ptr->orig_virt_addr = 0;
    }
#endif

    if (page_ptr->dma_addr != page_ptr->phys_addr)
    {
        pci_unmap_sg(dev, sg_ptr, 1, PCI_DMA_BIDIRECTIONAL);
        page_ptr->dma_addr = 0;
    }
}
#endif  /* NV_SG_MAP_BUFFERS */

/*
 * Basic support for kgdb assertions.
 */
#if defined(CONFIG_X86_REMOTE_DEBUG)
#include <linux/gdb.h>

#define NV_ASSERT(message, condition) KGDB_ASSERT(message, condition)
#else
#if defined(DEBUG)
#define NV_ASSERT(message, condition) \
if (!(condition)) { \
    nv_printf(NV_DBG_ERRORS, "NVRM: ASSERT: %s\n", message); \
    os_dbg_breakpoint(); \
}
#else
#define NV_ASSERT(message, condition)
#endif /* DEBUG */
#endif


#endif  /* _NV_LINUX_H_ */
